import json
import os
import socket
import bpy
from bpy.props import BoolProperty
from bpy.types import Context
import requests
import addon_utils

from pathlib import Path

from threading import Thread
from queue import Queue, Empty
import socketserver
from http.server import BaseHTTPRequestHandler
from time import sleep

from tempfile import NamedTemporaryFile

from typing import TypedDict


bl_info = {
    "name": "3Daistudio.com Blender Bridge",
    "author": "3daistudio/achillesdawn",
    "version": (1, 0, 0),
    "blender": (4, 3, 1),
    "description": "3Daistudio.com Blender Bridge",
    "category": "Tools",
}


class Config:
    normal_name = bl_info["name"]
    caps_name = "AISTUDIO"
    lower_case_name = caps_name.lower()
    addon_global_var_name = "ai_studio"

    path: Path

    @staticmethod
    def get_addon_name():
        addon_path = Path(__file__).resolve().parent.stem
        return addon_path


for mod in addon_utils.modules():  # type: ignore
    if mod.bl_info["name"] == bl_info["name"]:
        filepath = mod.__file__
        Config.path = Path(filepath).parent


class AISTUDIO_Properties(bpy.types.PropertyGroup):
    bridge_running: BoolProperty(default=False)  # type: ignore


class BaseClass:
    @staticmethod
    def get_props() -> AISTUDIO_Properties:
        return getattr(bpy.context.scene, Config.addon_global_var_name)  # type: ignore


class AISTUDIO_PT_viewportSidePanel(BaseClass, bpy.types.Panel):
    bl_idname = "AISTUDIO_PT_viewportSidePanel"
    bl_label = Config.normal_name
    bl_space_type = "VIEW_3D"
    bl_region_type = "UI"
    bl_context = "objectmode"
    bl_category = "3D AI Studio"

    def draw(self, context):
        layout = self.layout
        props = self.get_props()

        row = layout.row()

        if props.bridge_running:
            row.operator(
                AISTUDIO_OT_STOP.bl_idname,
                text="Bridge ON",
                icon="BLENDER",
                depress=True,
            )

        else:
            row.operator(
                AISTUDIO_OT_RECEIVE.bl_idname, text="Run Bridge", icon="BLENDER"
            )


class BaseOperator(BaseClass):
    @staticmethod
    def select_and_make_active(ob: bpy.types.Object):
        for ob_to_deselect in bpy.data.objects:
            if ob_to_deselect == ob:
                continue
            ob_to_deselect.select_set(False)

        assert bpy.context
        bpy.context.view_layer.objects.active = ob
        ob.select_set(True)

        print(f"[ Status ] {ob.name} set to Active Object")


class MeshTransfer(TypedDict):
    file_format: str
    path: str


queue: "Queue[MeshTransfer]" = Queue()

class BlenderMinimalServer(BaseHTTPRequestHandler):
    def _send_cors_headers(self):
        self.send_header('Access-Control-Allow-Origin', 'https://www.3daistudio.com')
        self.send_header('Access-Control-Allow-Methods', 'GET, POST, OPTIONS')
        self.send_header('Access-Control-Allow-Headers', '*')  # Allow any header
        self.send_header('Access-Control-Max-Age', '86400')  # 24 hours

    def do_OPTIONS(self):
        self.send_response(200)
        self._send_cors_headers()
        self.end_headers()

    def do_GET(self):
        print("📥 GET request:", self.path)
        print(self.headers)

        if self.path == "/status" or self.path == "/ping":
            self.send_response(200)
            self._send_cors_headers()
            self.send_header("Content-type", "application/json")
            self.end_headers()
            self.wfile.write(b'{"status": "ok"}')
        else:
            self.send_response(404)
            self._send_cors_headers()
            self.send_header("Content-type", "application/json")
            self.end_headers()
            self.wfile.write(b'{"status": "path not found"}')

    def do_POST(self):
        """Handle POST requests with URLs."""
        print("\n📥 POST request received:", self.path)
        print("Headers:", self.headers)

        if self.path == "/import":
            try:
                # Read content length
                content_length = int(self.headers.get('Content-Length', 0))
                print(f"Reading content length: {content_length}")

                # Read the body in smaller chunks
                chunks = []
                remaining = content_length
                while remaining > 0:
                    chunk_size = min(remaining, 1024)  # Read 1KB at a time
                    chunk = self.rfile.read(chunk_size)
                    if not chunk:
                        break
                    chunks.append(chunk)
                    remaining -= len(chunk)
                    print(f"Read {content_length - remaining}/{content_length} bytes")

                # Combine chunks and decode
                body = b''.join(chunks).decode('utf-8')
                print(f"Received body: {body}")
                
                # Parse JSON
                data = json.loads(body)
                glb_url = data.get('glb_url')
                print(f"Extracted URL: {glb_url}")

                if not glb_url:
                    raise ValueError("No GLB URL provided")

                # Download file
                print(f"Initiating download from {glb_url}")
                response = requests.get(glb_url, stream=True)
                response.raise_for_status()
                
                # Save to temp file
                temp_dir = os.path.join(os.path.expanduser('~'), 'AppData', 'Local', 'Temp')
                file_path = os.path.join(temp_dir, f"downloaded_model.glb")
                
                file_size = int(response.headers.get('content-length', 0))
                downloaded_size = 0

                with open(file_path, "wb") as f:
                    for chunk in response.iter_content(chunk_size=8192):
                        if chunk:
                            f.write(chunk)
                            downloaded_size += len(chunk)
                            if file_size:
                                print(f"Download progress: {(downloaded_size / file_size) * 100:.1f}%")

                print(f"Download complete: {file_path}")

                # Queue for import
                mesh_transfer: MeshTransfer = {
                    "file_format": "GLB",
                    "path": file_path
                }
                queue.put(mesh_transfer, block=False)

                # Send success response
                self.send_response(200)
                self._send_cors_headers()
                self.send_header("Content-type", "application/json")
                self.end_headers()
                self.wfile.write(json.dumps({
                    "status": "ok",
                    "message": "File queued for import"
                }).encode())

            except Exception as e:
                print(f"❌ Error: {str(e)}")
                import traceback
                traceback.print_exc()
                self.send_error_response(str(e))
        else:
            self.send_response(404)
            self._send_cors_headers()
            self.send_header("Content-type", "application/json")
            self.end_headers()
            self.wfile.write(b'{"status": "path not found"}')






_server_stop = False


def guard_job(server: socketserver.TCPServer):
    global _server_stop

    while not _server_stop:
        sleep(0.2)

    server.shutdown()

    print("joining guard")


def run_server():
    PORT = 7898
    with socketserver.TCPServer(("", PORT), BlenderMinimalServer) as server:
        guard = Thread(target=guard_job, args=[server])
        guard.start()

        print("[ Bridge ] listening on port 7898")
        server.serve_forever()

        server.server_close()
    print("joining server")


class AISTUDIO_OT_STOP(bpy.types.Operator, BaseOperator):
    bl_idname = Config.lower_case_name + ".stop"
    bl_label = Config.normal_name + " " + "Operator"
    bl_description = "Receives Payloads to load into blender"
    bl_options = {"REGISTER"}

    def execute(self, context: Context | None):
        props = self.get_props()
        props.bridge_running = False

        return {"FINISHED"}


class AISTUDIO_OT_RECEIVE(bpy.types.Operator, BaseOperator):
    bl_idname = Config.lower_case_name + ".receive"
    bl_label = Config.normal_name + " " + "Operator"
    bl_description = "Receives Payloads to load into blender"
    bl_options = {"REGISTER"}

    props: AISTUDIO_Properties
    _timer = None

    def process_mesh_transfer(self, transfer: MeshTransfer):
        print("🔄 Processing mesh transfer:", transfer)
        try:
            match transfer["file_format"]:
                case "GLB":
                    print(f"📥 Importing GLB from: {transfer['path']}")
                    result = bpy.ops.import_scene.gltf(filepath=transfer["path"])
                    print(f"✅ Import result: {result}")
                case _:
                    print(f"❌ Unsupported format: {transfer['file_format']}")
        except Exception as e:
            print(f"❌ Error processing mesh: {str(e)}")
        finally:
            # Clean up the temp file
            try:
                if os.path.exists(transfer["path"]):
                    os.remove(transfer["path"])
                    print(f"🗑️ Cleaned up temp file: {transfer['path']}")
            except Exception as e:
                print(f"❌ Error cleaning up: {str(e)}")

    def modal(self, context, event):
        if event.type in {"ESC"}:
            print("❌ ESC pressed, cancelling...")
            self.cancel(context)
            return {"CANCELLED"}

        if event.type == "TIMER":
            if self.props.bridge_running:
                try:
                    # Check queue size
                    qsize = queue.qsize()
                    if qsize > 0:
                        print(f"📦 Queue size: {qsize}")
                    
                    while True:
                        try:
                            item = queue.get(block=False)
                            print("📥 Got item from queue:", item)
                            self.process_mesh_transfer(item)
                            queue.task_done()
                        except Empty:
                            break
                except Exception as e:
                    print(f"❌ Error in modal timer: {str(e)}")
            else:
                print("⚠️ Bridge not running, cancelling...")
                self.cancel(context)
                return {"CANCELLED"}

        return {"PASS_THROUGH"}

    def execute(self, context):
        global _server_stop

        print("🚀 Starting bridge...")
        self.props = self.get_props()

        self.props.bridge_running = True
        _server_stop = False

        self.thread = Thread(target=run_server, daemon=True)
        self.thread.start()

        print("⚡ Server thread started")

        assert context
        wm = context.window_manager
        self._timer = wm.event_timer_add(0.1, window=context.window)
        wm.modal_handler_add(self)

        print("⏱️ Modal timer started")
        return {"RUNNING_MODAL"}

    def cancel(self, context):
        global _server_stop

        print("🛑 Cancelling bridge...")
        _server_stop = True
        try:
            print("🔄 Joining server thread...")
            self.thread.join()

            if self.thread.is_alive():
                raise Exception("Thread still alive")
        except Exception as e:
            print(f"❌ Error during shutdown: {str(e)}")

        print("🔄 Removing timer...")
        assert context
        wm = context.window_manager
        wm.event_timer_remove(self._timer)
        print("✅ Bridge shutdown complete")



classes = [
    AISTUDIO_OT_RECEIVE,
    AISTUDIO_OT_STOP,
    AISTUDIO_PT_viewportSidePanel,
    AISTUDIO_Properties,
]


def set_properties():
    setattr(
        bpy.types.Scene,
        Config.addon_global_var_name,
        bpy.props.PointerProperty(
            type=AISTUDIO_Properties,
        ),
    )


def del_properties():
    delattr(bpy.types.Scene, Config.addon_global_var_name)


def register():
    from bpy.utils import register_class

    for cls in classes:
        register_class(cls)

    set_properties()


def unregister():
    from bpy.utils import unregister_class

    for cls in reversed(classes):
        unregister_class(cls)

    del_properties()


if __name__ == "__main__":
    register()
